<?php
/**
 * Copyright (c) 2005-2008 GForge, LLC
 *
 * @version   $Id$
 * @author Tim Perdue tim@gforge.org
 * @date 2006-02-27
 *
 * This file is part of GForge.
 *
 * See LICENSE for details. This file is part of GForge AS and may not be
 * redistributed without written permission of GForge, LLC
 */

require_once('exceptions/InvalidFileException.php');
require_once('lib/StringUtils.php');

/**
 * This class parses an email with its headers
 */
class MailParser {
	const CONTENT_TYPE_TEXT = 'text/plain';
	const CONTENT_TYPE_HTML = 'text/html';
	
	private $lines = array();
	private $body;
	private $headers;
	private $from;
	private $content_type;
	private $is_valid = true;
	public $invalid_message = '';

	/**
	 * @param	mixed	If it's a string, it is the ile to read the email from. If it's an
	 *	 				array, it assumes it is the lines of the email.	If it's not defined, 
	 *					email is read from stdin.
	 */
	function __construct($input='php://stdin') {
		if (is_array($input)) {
			$this->lines = $input;
		} else {
			$this->lines = file($input);
			if ($this->lines === false) {
				throw new PermissionDeniedException($input);
			}
		}
		$this->parse();
	}
	
	private function parse() {
		/*$fh = fopen('/tmp/received-email', 'w');
		fputs($fh, implode('', $this->lines));
		fclose($fh);*/
		
		// First, parse the headers
		$i = 0;
		$lastheader = '';
		while (true) {
			$line = $this->lines[$i];
			// Empty line? Got to end of headers
			if (strlen(trim($line)) == 0) {
				$i++;
				break;
			}
			
			if ($i==0 && preg_match('/^From .+@.+/', $line)) {
				list($tmp, $value, $tmp2) = explode(' ', $line, 3);
				$header = 'From';
				$lastheader = $header;
				$value = trim($value);
				$this->headers[strtolower($header)] = $value;
			} else if (preg_match('/^[[:alnum:]]+.*:/', $line)) {
				list($header, $value) = explode(':', $line, 2);
				$header = trim($header);
				$lastheader = $header;
				$value = trim($value);
				$this->headers[strtolower($header)] = $value;
			} else if ($lastheader != '') {
				$line = trim($line);
				$this->headers[strtolower($lastheader)] .= $line;
				
			} else {
				// Don't know what happened... discarding data
			}
			
			$i++;
		}
		
		// We've finished parsing the headers. Now try to handle MIME stuff.
		if (array_key_exists('mime-version', $this->headers) && array_key_exists('content-type', $this->headers)) {
			$this->parseMIME($i);
		} else {
			// (thank god) It's not a MIME message. Read the whole body.
			$c = count($this->lines);
			for (;$i < $c; $i++) {
				$this->body .= $this->lines[$i];
			}
			$this->content_type = self::CONTENT_TYPE_TEXT;
		}
	}
	
	private function parseMIME($start_line) {
		$content_type = $this->headers['content-type'];
		$content_type_arr = explode(';', $content_type);
		
		$mime_content_type = strtolower(trim($content_type_arr[0]));
		if ($mime_content_type == self::CONTENT_TYPE_TEXT || $mime_content_type == self::CONTENT_TYPE_HTML) {
			$c = count($this->lines);
			// Regular text message... just read the body
			for ($i=$start_line; $i < $c; $i++) {
				$this->body .= $this->lines[$i];
			}
			$this->content_type = $mime_content_type;
			return;
		} else if ($mime_content_type == 'multipart/mixed' ||
					$mime_content_type == 'multipart/alternative' ||
					$mime_content_type == 'multipart/related' ||
					$mime_content_type == 'multipart/signed' ||
					$mime_content_type == 'multipart/plain') {
			// Worst case -- we have attachments
			// Find the boundary in the header
			$boundary = '';
			foreach ($content_type_arr as $content_type_part) {
				$content_type_part = trim($content_type_part);
				if (preg_match('/boundary="?([^"]*)"?/i', $content_type_part, $matches)) {
					$boundary = $matches[1];
					break;
				}
			}
			
			if (empty($boundary)) {
				// Error - couldn't find boundary
				$this->body = '[MailParser: couldn\'t parse MIME email] (1)';
				$this->invalid_message = $this->body;
				$this->is_valid = false;
			} else {
				$this->parseMultipartMessage($start_line, $boundary);
			}
		}
	}

	private function parseMultipartMessage($start_line, $boundary) {
		$separator = '--'.$boundary;
		$numlines = count($this->lines);
		
		// Skip the first part of the message, which is usually a text
		// informing the message is in MIME format
		$currline = $start_line;
		$line = '';
		while (true) {
			$line = trim($this->lines[$currline]);
			if ($line == $separator) {
				break;
			}
			$currline++;

			if ($currline == $numlines) {	// got to end of file
				$this->body = '[MailParser: couldn\'t parse MIME email] (2)';
				$this->invalid_message = $this->body;
				$this->is_valid = false;
				return;
			}
		}
		
		$parts = array();
		$currline++;
		$this->parseParts($currline, $separator, $parts);
		
		// Find the first valid (text or html) MIME part
		$found_message = false;
		for ($i=0; $i < count($parts); $i++) {
			$headers = $parts[$i]['headers'];
			foreach ($headers as $name => $value) {
				if (strtolower($name) == 'content-type') {
					// Found the Content-type header
					$content_parts = explode(';', $value);
					$content_type = strtolower($content_parts[0]);
					
					if ($content_type == self::CONTENT_TYPE_TEXT || $content_type == self::CONTENT_TYPE_HTML) {
						$this->content_type = $content_type;
						$this->body = $parts[$i]['content'];
						$found_message = true;
						break 2;		// Found the appropiate part. Stop parsing the parts.
					}
				}
			}
		}
		
		if (!$found_message) {
			$this->body = '[MailParser: couldn\'t parse MIME email] (3): No text content found.';
			$this->is_valid = false;
			$this->invalid_message = $this->body;
		}
	}
	
	private function parseParts($start_line, $separator, &$parts) {
		$currline = $start_line;
		$numlines = count($this->lines);
		while (true) {
			$part = array();
			$part['content'] = '';
			$part['headers'] = array();

			if ($currline == $numlines) {
				// Got to the end of file -- it means we've finished parsing the email
				return;
			}
			
			// Read the MIME header
			while (true) {
				$line = trim($this->lines[$currline]);
				if (empty($line)) break;		// End of header
				
				if (preg_match('/([^:]*): ?(.*)/', $line, $matches)) {
					$part['headers'][strtolower($matches[1])] = $matches[2];
				}
				
				$currline++;
				if ($currline >= $numlines) break 2;
			}
			
			$currline++;			// Move to the beginning of the message
			
			// Check if it's a multipart part (*sigh*)
			if (isset($part['headers']['content-type'])) {
				$content_type = $part['headers']['content-type'];
				$content_type_arr = explode(';', $content_type);
				$mime_content_type = strtolower(trim($content_type_arr[0]));
				if (strpos($mime_content_type, 'multipart') === 0) {
					// Yes, we have attachments inside the attachments!
					// Find out the separator
					$boundary = '';
					foreach ($content_type_arr as $content_type_part) {
						$content_type_part = trim($content_type_part);
						if (preg_match('/boundary="?([^"]*)"?/i', $content_type_part, $matches)) {
							$boundary = $matches[1];
							break;
						}
					}
					
					if (empty($boundary)) {
						// Error - couldn't find boundary
						$this->body = '[MailParser: couldn\'t parse MIME email] (4)';
						$this->invalid_message = $this->body;
						$this->is_valid = false;
					}
					$new_separator = '--'.$boundary;
					// Parse recursively
					$currline++;
					$processed_lines = $this->parseParts($currline, $new_separator, $parts);
					$currline += $processed_lines;
				}
			}

			// Now read the message until we get to the separator
			while (true) {
				$line = $this->lines[$currline];
				if (strpos($line, $separator) === 0) {
					// Separator found on the beginning of the line
					break;
				}
				
				$part['content'] .= $line;
				$currline++;
				if ($currline >= $numlines) break 2;
			}
			
			$part['content'] = trim($part['content']);
			$parts[] = $part;
			
			$currline++;

			// The separator with two extra dashes means it's the end of the MIME part
			if (strpos($line, $separator.'--') === 0) {
				break;
			}

			// End of message?
			if ($currline >= $numlines) break;
		}
		
		$processed_lines = ($currline-$start_line);
		return $processed_lines;
	}
	
	/**
	 * Returns false if the email had parsing errors
	 */
	public function isValid() {
		if (empty($this->headers['date']) || empty($this->headers['from']) ||
			empty($this->headers['message-id']) || empty($this->body)) {
			$this->is_valid = false;
			$this->invalid_message = 'Empty date, sender, message ID or body';
		}
		return $this->is_valid;
	}
	
	/**
	 * Returns the e-mail address of the sender
	 */
	public function getFrom() {
		$from = $this->headers['from'];
		
		if (preg_match('/[<]([^@]+@.+)[>]/', $from, $matches)) {
			$from = $matches[1];
		} else if (preg_match('/[(<]([^@]+@.+)[)>]/', $from, $matches)) {
			// $from is something like "John Doe" <john@doe.com>
			$from = $matches[1];
		} else if (preg_match('/([^@[:blank:]]+@[^[:blank:]]+)/', $from, $matches)) {
			// $from is something like:
			// * john@doe.com
			// * john@doe.com (johndoe)
			$from = $matches[1];
		} else {
			// couldn't parse the From header -- shouldn't happen
			$from = '';
		}
		
		return $from;
	}
	
	/**
	 * Returns the full "From:" string, not just the e-mail address
	 */
	public function getFullFrom() {
		return $this->headers['from'];
	}

	public function getBody() {
	 	if (!StringUtils::isUTF8($this->body)) {
			return (mb_convert_encoding(quoted_printable_decode($this->body),'UTF-8'));
		} else {
			return ($this->body);
		}
	}

	public function getSubject() {
		$subject = (isset($this->headers['subject'])) ? $this->headers['subject'] : '(no subject)';
		
		// Deal with subjects encoded in UTF8 (TODO: may have to do this for the body, or for other encodings?)
		if (preg_match_all('/=\?utf-8\?b\?(.*)=\?=/iU', $subject, $matches)) {
			for ($i = 0; $i < count($matches[1]); $i++) {
				$utf8string = $matches[1][$i];
				$string = base64_decode($utf8string);
				$subject = str_replace($matches[0][$i], $string, $subject);
			}
		}
		
		return $subject;
	}
	
	public function getMessageId() {
		if (isset($this->headers['message-id'])) {
			return $this->headers['message-id'];
		} else {
			return 0;
		}
	}
	
	/**
	 * Returns the UNIX timestamp of the date of the email
	 * @return	int
	 */
	public function getDate() {
		$date_str = $this->headers['date'];
		$timestamp = strtotime($date_str);
		return $timestamp;
	}
	
	public function getHeader($headerName)
	{
		$headerName = strtolower($headerName);
		if (!array_key_exists($headerName, $this->headers)) {
			return null;
		} else {
			return $this->headers[$headerName];
		}
	}
	
	public function getContentType() {
		return $this->content_type;
	}

}

?>
